#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <setjmp.h>
#include "config.h"

#ifdef READLINE
  #include <readline/readline.h>
  #include <readline/history.h>
#endif

typedef struct {
  ref a, b;
} cons;

typedef struct symbol {
  ref def;
  struct symbol *next;
  char name[1];
} symbol;

struct symval {
  ref r;
  symbol *sym;
};

struct fstack {
  struct fstack *next;
  ref fref;
  int symcnt;
  struct symval symbols[1];
};

#define MEMORY_BLOCK_SIZE 16384
#define ENOUGH_FREE_CONS 0x100
#define MAX_SYMBOL_NAME_LEN 256
#define SYMBOL_TABLE_LEN 256
#define LINE_LENGTH 2048

typedef struct mem_block {
  struct mem_block *next;
  cons *first, *last;
  char array[MEMORY_BLOCK_SIZE];
} mem_block;

mem_block *first_mem_block;
cons *first_avail_cons;
int forbid_garbage_collecting;
int avail_cons;
symbol *symbol_table[SYMBOL_TABLE_LEN];
FILE *inf;
int line;
jmp_buf env;
ref lambda_ref, special_ref, true_ref, rest_ref;
struct fstack *fstack;
int nesting;

#define get_int(val)      (int) (((sref)(val)) >> 3)
#define get_cons(val)     ((cons *) ((val) & ~(ref)7))
#define get_a(val)        (get_cons(val)->a)
#define get_b(val)        (get_cons(val)->b)
#define get_symbol(val)   ((symbol *) ((val) & ~(ref)7))
#define get_def(val)      (get_symbol(val)->def)
#define get_primitive(val) (int) ((val) >> 3)

#define make_int(val)     ((ref) ((sref)(val)) << 3)
#define make_cons(val)    ((ref) (val) + 1)
#define make_symbol(val)  ((ref) (val) + 3)
#define make_primitive(val) ( ((ref) (val) << 3) + 2) 
#define make_nil          make_primitive(0)
#define make_undef        make_primitive(1)
#define make_error        make_primitive(2)
#define make_eof          make_primitive(3)
#define make_close_paren  make_primitive(~(ref) 1)
#define make_dot          make_primitive(~(ref) 2)

#define is_int(val)       (((val) & 3) == 0)  
#define is_cons(val)      (((val) & 3) == 1)
#define is_symbol(val)    (((val) & 3) == 3)
#define is_primitive(val) (((val) & 3) == 2)
#define is_nil(val)       ((val) == make_nil) 
#define is_undef(val)     ((val) == make_undef)
#define is_eof(val)       ((val) == make_eof)
#define is_error(val)     ((val) == make_error)

#define PRIMITIVES P("nil",nil), P("undef",undef), P("error",error), \
		   P("",eof), P("+",plus), P("-",minus), P("*",mult), P("/",div), \
		   P("=",eqsign), P("<",lt), P(">",gt), P("<=",le), P(">=",ge), \
		   P("<>",ne), P("eq",eq), P("not",not), P("block",block), \
		   P("cons",cons), P("list",list), P("geta",geta), P("getb",getb), \
		   P("seta",seta), P("setb",setb), P("get",get), P("defined?",askdef),\
		   P("set",set), P("number?",asknum), P("pair?",askpair), \
		   P("symbol?",asksym), P("nil?",asknil), P("primitive?",askprim),\
		   P("eval",eval), P("if",if), P("and",and), P("or",or), \
		   P("quote",quote), P("define",define), P("let",let), \
		   P("garbage",garbage), P("print",print)

#define P(x,y) x
static char *primitives[] = { PRIMITIVES, NULL };
#undef P
#define P(x,y) P_##y
enum { PRIMITIVES, P_last } prim_values;
#undef P

#define F_PUSH(I,FREF) { \
      struct fstack *aux; \
      int _i = ((I) >= 0) ? (I) : 0; \
      aux = ST_ALLOC(sizeof(struct fstack) + (_i * sizeof(struct symval))); \
      aux->symcnt = _i; \
      aux->fref = FREF; \
      aux->next = fstack; \
      fstack = aux; \
    }

#define F_POP() { \
      int i; \
      struct fstack *aux = fstack; \
      i = aux->symcnt; \
      while (i-- > 0) \
         aux->symbols[i].sym->def = aux->symbols[i].r; \
      fstack = fstack->next; \
      ST_FREE(aux); \
    }

void print(ref);

void NONRET eval_error(char *mesg, ...) {
   va_list args;

   va_start(args, mesg);
   if (inf == stdin) 
      printf("Eval error: ");
   else
      printf("Eval error at line %d: ", line);
   vprintf(mesg, args);
   printf("\nStack trace:\n");

   while (fstack) {
     if (!is_nil(fstack->fref) && fstack->symcnt != -1) {
       print(fstack->fref);
       putchar('\n');
     }
     F_POP();
   }
   longjmp(env, 0);
}

void NONRET parse_error(char *mesg, ...) {
   va_list args;

   va_start(args, mesg);
   if (inf == stdin) 
      printf("Parse error: ");
   else 
      printf("Parse error at line %d: ", line);
   vprintf(mesg, args);
   putchar('\n');
   longjmp(env, 0);
}

void NONRET die(char *mesg, ...) {
   va_list args;

   va_start(args, mesg);
   printf("ksp lisp: ");
   vprintf(mesg, args);
   putchar('\n');
   exit(1);
}

void *xmalloc(int size) {
    void *res;
    if (!(res = malloc(size))) 
       die("Cannot allocate %d bytes of memory", size);
    return res;
}

void print(ref what) {
   if (is_int(what))
      printf("%d", get_int(what)); 
   else if (is_symbol(what)) 
      printf("%s", get_symbol(what)->name);
   else if (is_primitive(what))
      printf("#%s", primitives[get_primitive(what)]);
   else if (is_cons(what)) {
      ref aux;
      putchar('(');
      print(get_a(what));
      for (aux = get_b(what); is_cons(aux); aux = get_b(aux)) {
         putchar(' ');
         print(get_a(aux));
      }
      if (!is_nil(aux)) {
         printf(" . ");
         print(aux);
      }
      putchar(')');
   } 
}

void new_mem_block(void) {
   mem_block *res;
   cons *aux;
   res = xmalloc (sizeof (mem_block));
   memset(res, 0, sizeof(mem_block));
   res->next = first_mem_block;
   res->first = (cons *) (((ref)res->array + 7) & ~(ref)7);
   res->last = (cons *) ((ref)(res->array + MEMORY_BLOCK_SIZE - sizeof(cons)) &
~(ref)7);
   for (aux = res->first; aux < res->last; aux++)
      aux->a = (ref) (aux+1);
   aux->a = (ref) first_avail_cons;
   first_avail_cons = res->first;
   first_mem_block = res;
   avail_cons += res->last - res->first + 1;
}

#define QUEUED(REF) (get_a(REF) & (ref)4) 
#define CQUEUED(CONS) ((CONS)->a & (ref)4)
#define SEEN(REF)   (get_b(REF) & (ref)4)
#define CSEEN(CONS)   ((CONS)->b & (ref)4)
#define CUNQUEUE(CONS) do { (CONS)->a &= ~(ref)4; j--; } while (0)
#define CLEAR(CONS) do { (CONS)->a &= ~(ref)4; (CONS)->b &= ~(ref)4; } while (0)
#define QUEUE(REF)  if (is_cons(REF) && !SEEN(REF)) { \
                       get_a(REF) |= (ref)4; \
		       get_b(REF) |= (ref)4; \
		       j++; \
                    }

int garbage_collect(void) {
   int i,j = 0;
   struct fstack *f;
   symbol *sym;
   mem_block *mb;
   first_avail_cons = NULL;
   for (i = 0; i < SYMBOL_TABLE_LEN; i++) 
      for (sym = symbol_table[i]; sym; sym = sym->next)
         QUEUE(sym->def); 
   for (f = fstack; f; f = f->next) {
      QUEUE(f->fref);
      for (i = 0; i < f->symcnt; i++)
         QUEUE(f->symbols[i].r);
   }
   do {
      for (mb = first_mem_block; mb; mb = mb->next) {
          cons *cp;
          for (cp = mb->first; cp <= mb->last; cp++) {
             if (CQUEUED(cp)) {
	       CUNQUEUE(cp);
	       QUEUE(cp->a);
	       QUEUE(cp->b)
	     }
          }
      }
   } while (j > 0);
   j = 0;
   for (mb = first_mem_block; mb; mb = mb->next) {
      cons *cp;
      for (cp = mb->first; cp <= mb->last; cp++) {
         if (CSEEN(cp)) 
	    CLEAR(cp);
         else {
	    cp->a = (ref) first_avail_cons;
	    first_avail_cons = cp;
	    j++;
	 }
      }
   }
   return j;
}

ref new_cons(ref x, ref y) {
   cons *res;
   if (!first_avail_cons) {
      if (forbid_garbage_collecting || (avail_cons = garbage_collect()) <
ENOUGH_FREE_CONS)
         new_mem_block();
   }
   res = first_avail_cons;
   first_avail_cons = (cons*) (first_avail_cons->a);
   res->a = x;
   res->b = y;
   avail_cons--;
   return make_cons(res);
}

symbol *new_symbol(char *name) {
   unsigned int l = 0;
   char *s;
   symbol *sym;
   for (s = name; *s; s++)
       l = (l * 37) + *s;
   l %= SYMBOL_TABLE_LEN;
   for (sym = symbol_table[l]; sym; sym=sym->next) 
      if (!strcmp(sym->name, name)) return sym;
   sym = xmalloc (sizeof (symbol) + strlen(name) + 7);
   sym = (symbol *) ((((ref) sym) + 7) & ~(ref)7);
   strcpy(sym->name, name);
   sym->def = make_undef;
   sym->next = symbol_table[l];
   symbol_table[l] = sym; 
   return sym;
}

#define number(c) ((c)>='0' && (c)<='9')
#define white_space(c) ((c) == '\t' || (c) == ' ' || (c) == '\r' || \
						(c) == '\n')
#define unrecognized(c) ((c) == '\"' )
#define allowed_in_symbol_name(c) (!white_space(c) && !unrecognized(c) && \
								c != ')' && c != '(' && c !=EOF && c != '.')

void parse_init(void) {
    int i;
    symbol *sym;
    for (i = 0; primitives[i]; i++) {
       sym = new_symbol(primitives[i]);
       sym->def = make_primitive(i);
    }
    lambda_ref = make_symbol(new_symbol("lambda"));
    special_ref = make_symbol(new_symbol("special"));
    true_ref = make_symbol(new_symbol("t"));
    rest_ref = make_symbol(new_symbol("&rest"));
}

unsigned char *readline_s, *readline_p, *history_s = NULL;
char fgets_buf[LINE_LENGTH];

int FGETC(FILE *inf) {
   if (!readline_s) {
      line++;
      #ifdef READLINE
      if ((!nesting) && history_s) {
	 add_history(history_s); free(history_s); history_s = NULL;
      }
      #endif
      if (inf == stdin) 
      #ifdef READLINE
      {
         if (! (readline_p = readline_s = 
                 (unsigned char *) readline(nesting ? "? " : "> ")))
            return EOF;
      } else 
      #else
      printf(nesting ? "? " : "> ");
      #endif
      {
         if (! (readline_s = (unsigned char*) fgets(fgets_buf, LINE_LENGTH,
inf)))
	    return EOF;
	 readline_p = (unsigned char *)strchr((char *)readline_s, '\n');
	 if (!readline_p) {
            if (fgetc(inf) != -1)
	       parse_error("Line too long");
	 } else 
	    *readline_p = 0;
	 readline_p = readline_s;
      }
      #ifdef READLINE
      if (*readline_s) {
	 if (!history_s) {
	    history_s = xmalloc(strlen((char *)readline_s) + 1);
	    strcpy((char *)history_s, (char *)readline_s);
	 } else {
	    history_s = realloc(history_s, (strlen((char*)history_s) + strlen((char *)readline_s) + 2));
	    strcat((char *)history_s, (char *)"\n");
	    strcat((char *)history_s, (char *)readline_s);
         }
      }
      #endif
   }
   if (!*readline_p) {
      if ((char *)readline_s != fgets_buf)
          free(readline_s);
      readline_p = readline_s = NULL;
      return '\n';
   }
   return *(readline_p++);
}

void UNGETC(int what, FILE *inf) {
   if (readline_p < readline_s)
      die("You should never get here.");
   readline_p--;

   (void) what; (void) inf;
}

ref parse0(void) {
   int c, j = 0;
   char buf[MAX_SYMBOL_NAME_LEN+1], *end;
   ref res;
   sref i;

again:
   do { 
        c = FGETC(inf);
        if (c == '\n') continue;
   } while (white_space(c));
   if (unrecognized(c))
      parse_error ("Unrecognized character %c.", c);
   if (c == ')') return make_close_paren;
   if (c == EOF) return make_eof;
   if (c == '.') return make_dot;
   if (c == ';') { 
      do { c = FGETC(inf); } while (c != EOF && c != '\n');
      goto again;
   }
   if (c == '\'')  {
      ref x = new_cons(make_nil, make_nil);
      F_PUSH(0,x);
      get_a(x) = make_primitive(P_quote);
      get_b(x) = new_cons(make_nil, make_nil);
      get_a(get_b(x)) = parse0();
      F_POP();
      return x;
   }
   if (c == '(') {
     ref *rp;
     nesting ++;
     F_PUSH(-1,res = new_cons(make_nil, make_nil));
     rp = &(get_b(res));
     for(;;) {
       ref x = parse0();
       if (x == make_close_paren)
         break;
       if (x == make_eof)
         parse_error("End of file inside a list");
       if (x == make_dot) {
         x = parse0();
         if (x == make_close_paren || x == make_eof)
           parse_error(". must be followed by integer or symbol");
         *rp = x;
         if (parse0() != make_close_paren)
           parse_error(". must be followed by exactly one object");
         F_POP();
         nesting --;
         return get_b(res);
       }
       F_PUSH(0,x);
       *rp = x = new_cons(x, make_nil);
       F_POP();
       rp = &(get_cons(x)->b);
     }
     F_POP();
     *rp = make_nil;
     nesting --;
     return get_b(res);
   }
   while (allowed_in_symbol_name(c) && j < MAX_SYMBOL_NAME_LEN) {
     buf[j++] = c; c = FGETC(inf);
   }
   buf[j] = 0;
   if (j == MAX_SYMBOL_NAME_LEN)
      parse_error ("Symbol name %s too long. Max length is %d.", buf,
MAX_SYMBOL_NAME_LEN);
   UNGETC(c, inf);
   if (buf[0] != '+') {
      i = strtol(buf, &end, 10);
      if (!end || !*end) {
         if (i < -INT_MAX || i > INT_MAX)
            parse_error("Number too big %s", buf);
         else
            return make_int(i);
      } 
   }
   return make_symbol(new_symbol(buf));
}

ref parse(void) {
   ref res;
   if (setjmp(env))
      return make_error;
   if (!forbid_garbage_collecting && avail_cons < ENOUGH_FREE_CONS) 
      avail_cons = garbage_collect();
   forbid_garbage_collecting++;
   nesting = 0;
   res = parse0();
   if (res == make_close_paren)
      parse_error("Unmatched parenthesis");
   if (res == make_dot)
      parse_error("Misplaced .");
   forbid_garbage_collecting--;
   return res;
}

#define EVAL_INT(X) (is_int(k=eval(X)) ? get_int(k) : \
					(eval_error("Integer value expected"),0))

#define FOR_ALL(X,Y) \
           while (!is_nil(X)) { \
	      if (!is_cons(X)) \
		 { eval_error("Expecting list, got pair. \
Probably misplaced . or using cons instead of list"); } \
	      k = get_a(X); \
	      Y; \
	      X = get_b(X); \
	   }

#define FOR_ALL_EX(U,V,W,X,Y) { \
           if (is_nil(X)) U; \
	   if (!is_cons(X)) \
		  { eval_error("Expecting list, got pair. Probably misplaced \
 . or using cons instead of list"); } \
	   k = get_a(X); \
	   W; \
	   if (is_nil(X = get_b(X))) { V; } \
	   FOR_ALL(X,Y); }

#define FOR_ALL_EXI(U,V,X,Y) FOR_ALL_EX(U,V,i = EVAL_INT(k),X,Y)

#define IS_OF_TYPE(NAME,WHAT,TYPE) if (!is_cons(WHAT) || !is_nil(get_b(WHAT)))\
           { eval_error("%s requires one parameter", #NAME); }\
           return is_##TYPE(eval(get_a(WHAT))) ? true_ref : make_nil;

#define HAS_ANY_PAR(NAME,WHAT) if (!is_cons(WHAT) || !is_nil(get_b(WHAT)))\
           { eval_error("%s requires exactly one parameter", #NAME); }

#define HAS_ONE_PAR(NAME,WHAT,X,TYPE)  \
		   if (!is_cons(WHAT) || !is_##TYPE(X = eval(get_a(WHAT))) ||\
			   !is_nil(get_b(WHAT)) ) \
              { eval_error("%s requires one %s parameter",#NAME,#TYPE); }

#define HAS_TWO_PARS(NAME,WHAT,X,TYPE) \
		   if (!is_cons(WHAT) || !is_##TYPE(X = eval(get_a(WHAT))) ||\
			   !is_cons(WHAT = get_b(WHAT)) || !is_nil(get_b(WHAT)) ) \
 { eval_error("%s requires one %s and one other parameter",#NAME,#TYPE); }

ref eval(ref what);

ref eval_primitive(ref what, ref x) {
   sref i,k;
   ref y, *rp;
   what = get_b(what);
   switch(i = get_primitive(x)) {
      case P_nil :
      case P_undef :
      case P_error :
      case P_eof  :
              eval_error("Cannot execute %s", primitives[i]);
      case P_plus :
              i = 0;
	      FOR_ALL(what,i+=EVAL_INT(k));
              return make_int(i);
      case P_minus :
			  FOR_ALL_EXI(return make_int(0),return make_int(-i),what,
						  i-=EVAL_INT(k));
              return make_int(i);
      case P_mult :
              i = 1;
              FOR_ALL(what, i*=EVAL_INT(k));
              return make_int(i);
      case P_div :
			  FOR_ALL_EXI(return make_int(1),
			   eval_error("/ requires at least two parameters"),what,
			   i/=(k = EVAL_INT(k)) ? k : (eval_error("Division by zero"),0));
              return make_int(i);

      #if 0
      #define AUX(OP) { \
         ref i,j; \
	 if (!is_cons(what) || !is_int(i = eval(get_a(what))) || \
		 !is_cons(what = get_b(what))) \
	     eval_error("Two integer parameters required"); \
	 F_PUSH(-1,i); \
         if (!is_int(j = eval(get_a(what))) || !is_nil(get_b(what))) \
	        eval_error("Two integer parameters required"); \
         F_POP(); \
         return ((sref)i OP (sref)j) ? true_ref : make_nil; \
      }
      #endif
      #define AUX(Y) \
           x = true_ref; \
		   FOR_ALL_EXI(return true_ref,return true_ref,what,\
					   if (!(i Y (k = EVAL_INT(k)))) x = make_nil; i = k;) \
					 return x;
      case P_eqsign : AUX(==)
      case P_lt : AUX(<)
      case P_gt : AUX(>)
      case P_le : AUX(<=)
      case P_ge : AUX(>=)
      case P_ne : AUX(!=)
      #undef AUX
      case P_eq : y = true_ref;
				  FOR_ALL_EX(return true_ref,,x = eval(k),what,
							 if(x != eval(k)) y = make_nil);
	          return y;
      case P_not : HAS_ANY_PAR(not,what);
		   return is_nil(eval(get_a(what))) ? true_ref : make_nil;
      case P_block : x = make_nil;
                     FOR_ALL(what,x = eval(k));
                     return x;
      case P_cons : if (!is_cons(what) || !is_cons(x = get_b(what)) ||
						!is_nil(get_b(x)))
                       eval_error("cons requires exactly 2 parameters");
		    F_PUSH(-1,y = new_cons(make_nil, make_nil));
		    get_a(y) = eval(get_a(what));
		    get_b(y) = eval(get_a(x));
		    F_POP();
                    return y;
      case P_list : rp = &x;
                    *rp = make_nil;
		    FOR_ALL_EX(return make_nil, F_POP(); return x, 
			 F_PUSH(-1, *rp = new_cons(make_nil,make_nil));
			 get_a(*rp) = eval(k);
		         rp = &(get_b(*rp));,
			 what,F_PUSH(-1,*rp = new_cons(make_nil,make_nil));
 	                 get_a(*rp) = eval(k);
			 F_POP();
		         rp = &(get_b(*rp)));
		    F_POP();
		    return x;
      case P_geta : HAS_ONE_PAR(geta,what,x,cons);
		    return get_a(x);
      case P_getb : HAS_ONE_PAR(getb,what,x,cons);
		    return get_b(x);
      case P_seta : HAS_TWO_PARS(seta,what,x,cons); 
                    F_PUSH(-1,x);
                    get_a(x) = eval(get_a(what));
		    F_POP();
		    return x;
      case P_setb : HAS_TWO_PARS(setb,what,x,cons); 
                    F_PUSH(-1,x);
                    get_b(x) = eval(get_a(what));
		    F_POP();
		    return x;
      case P_get : HAS_ONE_PAR(get,what,x,symbol);
				   return (is_undef(x = get_def(x))) ?
					(eval_error("Dereference of an undefined symbol"),0) : x;
      case P_askdef : HAS_ONE_PAR(defined?,what,x,symbol);
                      return (is_undef(get_def(x)) ? make_nil : true_ref);
      case P_set : HAS_TWO_PARS(set,what,x,symbol);
                   F_PUSH(-1,x);
                   get_def(x) = eval(get_a(what));
		   F_POP();
		   return get_def(x);
      case P_asknum : IS_OF_TYPE(number?,what,int);
      case P_askpair : IS_OF_TYPE(pair?,what,cons);
      case P_asksym : IS_OF_TYPE(symbol?,what,symbol);
      case P_asknil : IS_OF_TYPE(nil?,what,nil);
      case P_askprim : IS_OF_TYPE(primitive?,what,primitive);
      case P_eval : HAS_ANY_PAR(eval,what);
                    F_PUSH(-1,y = eval(get_a(what)));
		    y = eval(y);
		    F_POP();
		    return y;
	  case P_if :  if (!is_cons(what) || !is_cons(x = get_b(what)) ||\
					   !is_cons(x = get_b(x)) || !is_nil(get_b(x)))
                  eval_error("if requires three parameters");
		  return (!is_nil(eval(get_a(what)))) ? eval(get_a(get_b(what))) :
				  eval(get_a(x));
      case P_and  : x = true_ref;
                    FOR_ALL(what,if(is_nil(x = eval(k))) return make_nil);
                    return x;
      case P_or :  FOR_ALL(what,if(!is_nil(x = eval(k))) return x);
                   return make_nil;
      case P_quote : HAS_ANY_PAR(quote,what); 
                     return get_a(what);
      case P_define : if (!is_cons(what) || (!is_cons(x = get_a(what)) && 
!is_symbol(x)) || !is_cons(what = get_b(what))) 
eval_error("define needs one cons or symbol and at least one other parameter");
	  if (is_symbol(x))
	     return (get_def(x) = get_a(what));
	  else {
	     symbol *sym = is_symbol(get_a(x)) ? get_symbol(get_a(x)) :
					   (eval_error("Symbol expected"),(symbol*)NULL);
	     F_PUSH(-1, y = new_cons(lambda_ref, make_nil));
	     get_b(y) = new_cons(get_b(x), what);
	     F_POP();
	     return sym->def = y;
	  }

      case P_let : if (!is_cons(what) || !is_cons(x = get_a(what)))
                       eval_error("let requires one list parameter");
          i = 0;
	  FOR_ALL(x, i++);
	  F_PUSH(i,make_nil);
	  fstack->symcnt = i = 0; x = get_a(what);
	  FOR_ALL(x, if (!is_cons(k) || !is_symbol(y = get_a(k))) 
	                  eval_error("Symbol - value pairs expected");
		     fstack->symbols[i].sym = get_symbol(y);
		     fstack->symbols[i].r = fstack->symbols[i].sym->def;
		     fstack->symbols[i].sym->def = is_cons(y = get_b(k)) ? 
										   eval(get_a(y)) : eval(y);
		     i++;
		     fstack->symcnt++;
	  );
	  y = make_nil;
	  x = get_b(what);
	  FOR_ALL(x, y = eval(k));
	  F_POP();
	  return y;
      case P_garbage: if (!is_nil(what))
                         eval_error("Garbage primitive takes no parameters");
		      printf("Running garbage collector\n Available cons: ");
                      return make_int(avail_cons = garbage_collect());
      case P_print: i = 0;
                    FOR_ALL(what, print(eval(k)); putchar(' '); i++);
		    putchar('\n');
                    return make_int(i);
      default : 
               eval_error("Unknown primitive %d.", i);
   }
   return make_error;
}

ref eval_function(ref what, ref x, int lambda) {
   ref y, *rp;
   sref k;
   int i, rest = 0;

   if (is_cons(y = get_a(get_b(x)))) {
      i = 0; 
      FOR_ALL(y, i++);
      F_PUSH(i,make_nil);
      fstack->symcnt = i = 0; y = get_a(x = get_b(x));
      FOR_ALL(y, if (!is_symbol(k)) {
                    F_PUSH (0,eval(get_a(what)));
eval_error("Incorrectly defined function. List of symbols expected as params.");
				 }
      if ((ref) k == rest_ref) {
         if (!is_cons (y = get_b(y)))
	    eval_error("Symbol name expected after &rest");
	 if (!is_nil (get_b(y)))
	    eval_error("Only one symbol expected after &rest");
	 k = get_a(y);
         fstack->symbols[i].sym = get_symbol(k);
	 rp = &(fstack->symbols[i].r);
	 what = get_b(what);
	 *rp = make_nil;
         FOR_ALL(what, *rp = new_cons(make_nil, make_nil); 
                       get_a(*rp) = lambda ? eval(k) : k;
	               rp = &(get_b(*rp))); 
	 rest = 1;
	 fstack->symcnt++;
      } else {
         fstack->symbols[i].sym = get_symbol(k);
         if (is_cons(what = get_b(what)))
           fstack->symbols[i].r = lambda ? eval(get_a(what)) :
	                                   get_a(what);
         else
	   eval_error("Too few parameters");
	 i++;
	 fstack->symcnt++;
      }
      );
      if (!rest && !is_nil(get_b(what)))
        eval_error("Too many parameters"); 
      for (i = 0; i < fstack->symcnt; i++) {
	ref r = fstack->symbols[i].sym->def;
	fstack->symbols[i].sym->def = fstack->symbols[i].r;
	fstack->symbols[i].r = r;
      }
      y = make_nil;
      x = get_b(x);
      FOR_ALL(x, y = eval(k));
      F_POP(); 
      return y;
   }
   y = make_nil;
   x = get_b(get_b(x));
   FOR_ALL(x, y = eval(k)); 
   return y;
}

ref eval(ref what) {
  ref x;
  if (is_cons(what)) {
      F_PUSH(0,what);
      x = eval(get_a(what));
      if (is_primitive(x)) {
         what = eval_primitive(what, x); 
      } else if (is_cons(x)) {
         if (get_a(x) == lambda_ref) {
	    what = eval_function(what,x,1);
	 } else if (get_a(x) == special_ref) {
	    what = eval_function(what,x,0);
	 } else
           eval_error("lambda or special expected here"); 
      } else if (is_symbol(x))
         eval_error("Cannot execute %s", get_symbol(x)->name);
      else
         eval_error("Cannot execute %d", get_int(x));
      F_POP();
  } else if (is_symbol(what))
	  what = is_undef(x = get_def(what)) ?
   	     eval_error("Dereference of an undefined symbol %s",
		  get_symbol(what)->name), 0  : x;
  return what;
}

ref eval0(ref what) {
  if (setjmp(env))
      return make_error;
  return eval(what); 
}

int main(int argc, char **argv) {
   int i;
   ref res;
   parse_init();
   forbid_garbage_collecting = 0;
   for (i = 1; i < argc; i++)  {
      line = 0;
      if (!(inf = fopen(argv[i], "r"))) 
         die("Cannot open input file %s", argv[i]);
      do {
         res = eval0(parse());
         if (is_error(res)) {
	   printf("in file %s\n",argv[i]);
	   return 1;
	 }
      } while (!is_eof(res)); 
      fclose(inf);
   }
   inf = stdin;
   while(!is_eof(res = parse())) {
      print(eval0(res)); 
      putchar('\n');
   }
   putchar('\n');
   return 0;
}

